#ifndef MEMORY_H
#define MEMORY_H 1

 /*
  * UAE - The Un*x Amiga Emulator
  *
  * memory management
  *
  * Copyright 1995 Bernd Schmidt
  * vim:ts=8:sw=4:
  */

extern void memory_reset (void);

extern int special_mem;
extern int display_mem;
#define S_READ 1
#define S_WRITE 2

typedef uae_u32 (*mem_get_func)(uaecptr) REGPARAM;
typedef void (*mem_put_func)(uaecptr, uae_u32) REGPARAM;
typedef uae_u8 *(*xlate_func)(uaecptr) REGPARAM;
typedef int (*check_func)(uaecptr, uae_u32) REGPARAM;

extern char *address_space, *good_address_map;
extern uae_u8 *chipmemory;

extern uae_u32 allocated_chipmem;
extern uae_u32 allocated_fastmem;
extern uae_u32 allocated_bogomem;
extern uae_u32 allocated_gfxmem;
extern uae_u32 allocated_z3fastmem;
extern uae_u32 allocated_a3000mem;

#include "machdep/maccess.h"

#if defined(USE_COMPILER) && !defined(USE_MAPPED_MEMORY)
#define USE_MAPPED_MEMORY
#endif

#define kickmem_size 0x080000
#define erommem_size 0x080000

#define chipmem_start 0x00000000
#define bogomem_start 0x00C00000
#define a3000mem_start 0x07000000
#define kickmem_start 0x00F80000
#define erommem_start 0x00F00000

extern int ersatzkickfile;
extern int cloanto_rom;

extern uae_u8* baseaddr[];

typedef struct {
    /* These ones should be self-explanatory... */
    mem_get_func lget, wget, bget;
    mem_put_func lput, wput, bput;
    /* Use xlateaddr to translate an Amiga address to a uae_u8 * that can
     * be used to address memory without calling the wget/wput functions.
     * This doesn't work for all memory banks, so this function may call
     * abort(). */
    xlate_func xlateaddr;
    /* To prevent calls to abort(), use check before calling xlateaddr.
     * It checks not only that the memory bank can do xlateaddr, but also
     * that the pointer points to an area of at least the specified size.
     * This is used for example to translate bitplane pointers in custom.c */
    check_func check;
    /* For those banks that refer to real memory, we can save the whole trouble
       of going through function calls, and instead simply grab the memory
       ourselves. This holds the memory address where the start of memory is
       for this particular bank. */
    uae_u8 *baseaddr;
    const char *name;
} addrbank;

extern uae_u8 *filesysory;
extern uae_u8 *rtarea;

extern addrbank dummy_bank;
extern addrbank chipmem_bank;
extern addrbank kickmem_bank;
extern addrbank custom_bank;
extern addrbank clock_bank;
extern addrbank cia_bank;
extern addrbank rtarea_bank;
extern addrbank expamem_bank;
extern addrbank fastmem_bank;
extern addrbank gfxmem_bank;
extern addrbank gfxcard_bank;
extern addrbank wdc_bank;
extern addrbank lance_bank;

extern void rtarea_init (void);
extern void rtarea_setup (void);
extern void expamem_init (void);
extern void expamem_reset (void);

extern uae_u32 gfxmem_start;
extern uae_u8 *gfxmemory;
extern uae_u32 gfxmem_mask;
extern int address_space_24;


/* Default memory access functions */
extern int default_check(uaecptr addr, uae_u32 size) REGPARAM;
extern uae_u8 *default_xlate(uaecptr addr) REGPARAM;

#define bankindex(addr) (((uaecptr)(addr)) >> 16)

extern addrbank *mem_banks[65536];
extern uae_u8 *baseaddr[65536];

#define get_mem_bank(addr) (*mem_banks[bankindex(addr)])
#define put_mem_bank(addr, b, realstart) do { \
    (mem_banks[bankindex(addr)] = (b)); \
    if ((b)->baseaddr) \
        baseaddr[bankindex(addr)] = (b)->baseaddr - (realstart); \
    else \
        baseaddr[bankindex(addr)] = (uae_u8*)(((long)b)+1); \
} while (0)

extern void memory_init (void);
extern void memory_cleanup (void);
extern void map_banks (addrbank *bank, int first, int count, int realsize);

#ifndef MD_HAVE_MEM_1_FUNCS

STATIC_INLINE uae_u32 longget(uaecptr addr) {
	return call_mem_get_func(get_mem_bank(addr).lget, addr);
}
STATIC_INLINE uae_u16 wordget(uaecptr addr) {
	return call_mem_get_func(get_mem_bank(addr).wget, addr);
}
STATIC_INLINE uae_u8 byteget(uaecptr addr) {
	return call_mem_get_func(get_mem_bank(addr).bget, addr);
}
STATIC_INLINE void longput(uaecptr addr, uae_u32 l) {
	call_mem_put_func(get_mem_bank(addr).lput, addr, l);
}
STATIC_INLINE void wordput(uaecptr addr, uae_u16 w) {
	call_mem_put_func(get_mem_bank(addr).wput, addr, w);
}
STATIC_INLINE void byteput(uaecptr addr, uae_u8 b) {
	call_mem_put_func(get_mem_bank(addr).bput, addr, b);
}

#define longget_1 longget
#define wordget_1 wordget
#define byteget_1 byteget
#define longput_1 longput
#define wordput_1 wordput
#define byteput_1 byteput

#endif

#if 1 /* MMU code */

#define get_byte mmu_get_byte
#define get_word mmu_get_word
#define get_long mmu_get_long
#define put_byte mmu_put_byte
#define put_word mmu_put_word
#define put_long mmu_put_long

#else

STATIC_INLINE uae_u32 get_long(uaecptr addr) {
	return longget_1(addr);
}
STATIC_INLINE uae_u32 get_word(uaecptr addr) {
	return wordget_1(addr);
}
STATIC_INLINE uae_u32 get_byte(uaecptr addr) {
	return byteget_1(addr);
}
STATIC_INLINE void put_long(uaecptr addr, uae_u32 l) {
	longput_1(addr, l);
}
STATIC_INLINE void put_word(uaecptr addr, uae_u32 w) {
	wordput_1(addr, w);
}
STATIC_INLINE void put_byte(uaecptr addr, uae_u32 b) {
	byteput_1(addr, b);
}

#endif

STATIC_INLINE uae_u8 *phys_get_real_address(uaecptr addr)
{
    return get_mem_bank(addr).xlateaddr(addr);
}

STATIC_INLINE int phys_valid_address(uaecptr addr, uae_u32 size)
{
    return get_mem_bank(addr).check(addr, size);
}

/* For faster access in custom chip emulation.  */
extern uae_u32 chipmem_lget (uaecptr) REGPARAM;
extern uae_u32 chipmem_wget (uaecptr) REGPARAM;
extern uae_u32 chipmem_bget (uaecptr) REGPARAM;
extern void chipmem_lput (uaecptr, uae_u32) REGPARAM;
extern void chipmem_wput (uaecptr, uae_u32) REGPARAM;
extern void chipmem_bput (uaecptr, uae_u32) REGPARAM;

#ifdef NATMEM_OFFSET

typedef struct shmpiece_reg {
    uae_u8 *native_address;
    int id;
    uae_u32 size;
    struct shmpiece_reg *next;
    struct shmpiece_reg *prev;
} shmpiece;

extern shmpiece *shm_start;
extern int canbang;

#endif

extern uae_u8 *mapped_malloc (size_t, char *);
extern void mapped_free (uae_u8 *);

#endif
